'use strict';

/**
 * each: Designed to loop over a collection, Array or Object, and applies the action 
 * Function to each value in the collection.
 * 
 * @param {Array or Object} collection The collection over which to iterate.
 * @param {Function} action The Function to be applied to each value in the 
 * collection
 */
function each(collection, action) {
    if (Array.isArray(collection)) {
        for (var i = 0; i < collection.length; i++) {
            action(collection[i], i, collection);
        }
    }
    else {
        for (var key in collection) {
            action(collection[key], key, collection);
        }
    }
}
module.exports.each = each;

//have to export higher order functions like this ^
/**
 * typeOf: 
 * 
 * @param {value} 
 * @return {value} 
 */
function typeOf(value) {
    if (Array.isArray(value)) return 'array';
    if (value === null) return 'null';
    if (value instanceof Date) return 'date';
    return typeof value;
}
module.exports.typeOf = typeOf;

/**
 * filter: what does filter do???
 * 
 * @param {Array or Object} collection  An Array or Object or whose values will be filtered.
 * @param {Function} test A Funtion that takes  the value, position and collection for each value in collection, 
 * and returns true if the value passes some test, false otherwise.
 * @return {}
 * Usage:
 * 
 * var p = [{name: 'John'}, {name: 'Max'}] 
 * example of test:
 * 
 * var startsWithJ = function(person, i, people) {
 *    return person.name[0] === 'J';
 * }
 *
 * var allJPeople = filter(p, startsWithJ);
 */
function filter(collection, test) {
    var results = [];
    each(collection, function(value, position, collection) {
        if (test(value, position, collection)) results.push(value);
    });
    return results;
}
module.exports.filter = filter;

/* _.identity()
 * What does it do?  
 * It returns the value of whatever was passed into it
 * Examples:
 *   _.identity(5) === 5
 *   _.identity({a: "b"}) === {a: "b"}
 */
function identity(value) {
    return value;
}
module.exports.identity = identity;


/** _.first()
What does it do?

_.first() takes an array and a number as parameters
and returns the first item or items in the array;
the number of items in the returned array is determined by the number parameter;
if the first parameter that is suppose to be an array is NOT an array, an empty 
array is returned.  if a number that is suppose to be passed in the number parameter is not given
or not a number, then the first element of the array is returned.

 * Examples:
 *   _.first(["a","b","c"], 1) -> "a"
 *   _.first(["a","b","c"], 2) -> ["a", "b"]
 *   _.first(["a", "b", "c"], "ponies") -> ["c"]
 *  note to self:  YOU NEED TO WORK ON THIS CODE SOME MORE
 *                 A STRING PASSED AS NUMBER RETURNS 'undefined'*/

function first(array, n) {
    if (!Array.isArray(array) || n < 0) return [];
    if (n === undefined) return array[0];

    if (n > 0) return array.slice(0, n);
    if (n < 0) {
        n = n > array.length ? array.length : n;
        return array.slice(0, n);
    }
}
module.exports.first = first;


/** _.last()
*  What does it do?
* _.last() takes an array and a number as parameters and returns the elements of the array
* that are at the end of the array based off of the numbers parameter.  if the array parameter
* is not an array, an empty array is returned;  if the number parameter is not included or is not
* a number, then the last element of the array is returned.
*
** Examples:
*   _.last(["a","b","c"], 2) -> ["b","c"]
*   _.last(["a", "b", "c"], "ponies") -> ["c"]
*note to self:  YOU NEED TO WORK ON THIS CODE SOME MORE
 *                 A STRING PASSED AS NUMBER RETURNS 'undefined'
                      INSTEAD OF LAST ELEMENT*/

function last(array, n) {
    if (!Array.isArray(array) || n < 0) return [];
    if (n === undefined) return array.pop();
    if (n < 0) return [];
    if (array.length > n > 0) return array.slice(-2);
    if (n >= array.length) return array;
}
module.exports.last = last;


/** _.indexOf()
 * What does it do?
 * _.indexOf() takes an array as it's first parameter and some type of value as it's 
 * second parameter, the function loops over the array and returns the index of the 
 * array where the value is first found.  if the value is not found in the array, -1 is returned.
 *   _.indexOf(["a","b","c"], "c") -> 2
 *   _.indexOf(["a","b","c"], "d") -> -1
 * note to self:  NEED TO GO BACK AND FIGURE OUT HOW TO WRITE THIS WITHOUT USING
 *     INDEX OF
 */
function indexOf(array, value) {
    for (var i = 0; i < array.length; i++) return array.indexOf(value);
}
module.exports.indexOf = indexOf;


/** _.reject()
 * What does it do?
 * _.reject takes and array and a function as parameters, the function is called for 
 * each element in the array and returns an array of elements of which the function
 * returned FALSE
 * 
 * Examples : 
 *   _.reject([1,2,3,4,5], function(e){return e%2 === 0}) -> [1,3,5]
 */
function reject(obj, predicat, context) {
    var results = [];
    filter(obj, function(value, index, list) {
        if (!predicat(value, index, list)) results.push(value);
    });
    return results;
}
module.exports.reject = reject;

/** _.partition()
* What does it do?  
*  _.partition takes an array and a function as parameters
* the function is called for each element in the array passing : element, key,and
* array as the arguments;  an array is returned based on whether the function returns
* something that is true or false
*    
* Examples:
*   _.partition([1,2,3,4,5], function(element,index,arr){
*     return element % 2 === 0;
*   }); -> [[2,4],[1,3,5]]
}
*/
function partition(obj, predicate, context) {
    predicate = identity(predicate, context);
    var pass = [],
        fail = [];
    each(obj, function(value, key, obj) {
        (predicate(value, key, obj) ? pass : fail).push(value);
    });
    return [pass, fail];
}
module.exports.partition = partition;

/*_unique()
* What does it do?
* _.unique() takes an array as a parameter and returns an array that has any 
elements that are duplicated removed


*/

function unique(a) {
    var prims = {
            "boolean": {},
            "number": {},
            "string": {}
        },
        objs = [];
    return a.filter(function(item) {
        var type = typeof item;
        if (type in prims)
            return prims[type].hasOwnProperty(item) ? false : (prims[type][item] = true);
        else
            return objs.indexOf(item) >= 0 ? false : objs.push(item);
    });
}
module.exports.unique = unique;



/*_.map()
 * What does it do?
 * _.map takes a collection (either an array or an object) and a function as parameters,
 * returns a new array of values by mapping each value in list through a transformation function (iteratee).
 * the iteratee is passed three arguments: the value, then the index (or key) of the iteration, 
 * and finally a reference to the entire list.
 *
 */

function map(obj, iteratee, context) {
    iteratee = identity(iteratee, context);
    var keys = !Array.isArray(obj) && Object.keys(obj),
        length = (keys || obj).length,
        results = Array(length);
    for (var index = 0; index < length; index++) {
        var currentKey = keys ? keys[index] : index;
        results[index] = iteratee(obj[currentKey], currentKey, obj);
    }
    return results;
}
module.exports.map = map;

/** _.contains()
 *  What does it do?
 * 
 * _.contains takes an array and an item as parameters,  if the item is one of the elements, 
 * true is returned, if the item is NOT one of the elements in the arrray, 
 * false is returned
 * var arr = [1,2,3,4,5];
 * var truth = contains(arr, 5); // true
 * var truth = contains(arr, 7); // false
 * return truth;
 * http://jsbin.com/hejoru/2/edit?html,js,output
 * */

function contains(obj, item) {
    return (obj).includes(item);
}
module.exports.contains = contains;

/** _.every()
*  What does it do?
* _.every() takes an collection (either an array or an object) as parameter and also takes
* a function as a parameter. if the collection is an array, the current element, it's index
* and the array is passed to the function;  if the collection is an object,  it's current value,
* current key, and object is passed to the function,  if the RETURN value of the calling 
* function returns TRUE, it returns TRUE,  if any one of the returning values comes back FALSE, return false
* 


* Examples:
*   _.every([2,4,6], function(e){return e % 2 === 0}) -> true
*   _.every([1,2,3], function(e){return e % 2 === 0}) -> false
* _.every() pretty much works like contains except!! instead of a value being passed
* as the second parameter, a real live function is passed;
*/
function every(obj, predicate) {
    predicate = predicate || identity;
    var keys = !Array.isArray(obj) && Object.keys(obj),
        length = (keys || obj).length;
    for (var index = 0; index < length; index++) {
        var currentKey = keys ? keys[index] : index;

        if (predicate === !true || false) return true;
        if (!predicate(obj[currentKey], currentKey, obj)) return false;
    }
    return true;
}
module.exports.every = every;

/** _.some()
 * What does it do? 
 * _.some() takes a collection(either an array or an object) & a function as parameter
 * if the collection is an array, the current element, its index, and array get passed to the function
 * if the collection is an object, the current value, current key and object get passed;
 * if at least one the returning elements from the function is true, then true is returned
 * only return false if EVERY returning element is false
 * 
 * Examples:
 *   _.some([1,3,5], function(e){return e % 2 === 0}) -> false
 *   _.some([1,2,3], function(e){return e % 2 === 0}) -> true
 * */
function some(obj, predicate) {
    predicate = predicate || identity;
    var keys = !Array.isArray(obj) && Object.keys(obj),
        length = (keys || obj).length;
    for (var index = 0; index < length; index++) {
        var currentKey = keys ? keys[index] : index;
        if (predicate === "") return false;
        if (predicate(obj[currentKey], currentKey, obj)) return true;
    }
    return false;
}
module.exports.some = some;

/** _.extend 
* What does it do?
* _.extend takes any number of objects as parameters, adds all the values of the 
* objects together for one big happy object
* 
* examples :
* var data = {a:"one"};
*   _.extend(data, {b:"two"}); -> data now equals {a:"one",b:"two"}
*   _.extend(data, {a:"two"}); -> data now equals {a:"two"}
* 
* */
function extend(obj){
    for (var i = 1;i < arguments.length; i++){
        each(arguments[i], function(value,keys,collection){
            obj[keys] = value;
        });
    }
    return obj;
}   

module.exports.extend = extend;

/*takes an array and a property as parameters
/* Returns an array containing the value of <property> for every element in <array>
/* Examples:
     *   _.pluck([{a: "one"}, {a: "two"}], "a") -> ["one", "two"]
     */
function pluck(obj, key) {
    return map(obj, function(x) {
        return x[key];
    });
}
module.exports.pluck = pluck;


/*Arguments:
*   1) An array
*   2) A function
*   3) A seed
* Reduces collection to a value which is the accumulated result of running each 
* element in collection thru iteratee, where each successive invocation is supplied 
* the return value of the previous. If accumulator is not given, the first element of 
* collection is used as the initial value */



function reduce(array, action, seed) {
    var prev = seed;
    for(var i = 0; i < array.length; i++) {
        if(prev === undefined) {
            prev = array[0];
        } else {
            var next = action(prev, array[i], i);
            prev = next;
        }
    }
    return prev;
  }  
module.exports.reduce = reduce;





